@@ -4,7 +4,7 @@
  *	  vacuuming routines for the postgres GiST index access method.
  *
  *
- * Portions Copyright (c) 1996-2013, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2014, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -34,7 +34,6 @@ gistvacuumcleanup(PG_FUNCTION_ARGS)
 	BlockNumber npages,
 				blkno;
 	BlockNumber totFreePages;
-	double		tuplesCount;
 	bool		needLock;
 
 	/* No-op in ANALYZE ONLY mode */
@@ -43,7 +42,17 @@ gistvacuumcleanup(PG_FUNCTION_ARGS)
 
 	/* Set up all-zero stats if gistbulkdelete wasn't called */
 	if (stats == NULL)
+	{
 		stats = (IndexBulkDeleteResult *) palloc0(sizeof(IndexBulkDeleteResult));
+		/* use heap's tuple count */
+		stats->num_index_tuples = info->num_heap_tuples;
+		stats->estimated_count = info->estimated_count;
+
+		/*
+		 * XXX the above is wrong if index is partial.  Would it be OK to just
+		 * return NULL, or is there work we must do below?
+		 */
+	}
 
 	/*
 	 * Need lock unless it's local to this backend.
@@ -58,7 +67,6 @@ gistvacuumcleanup(PG_FUNCTION_ARGS)
 		UnlockRelationForExtension(rel, ExclusiveLock);
 
 	totFreePages = 0;
-	tuplesCount = 0;
 	for (blkno = GIST_ROOT_BLKNO + 1; blkno < npages; blkno++)
 	{
 		Buffer		buffer;
@@ -76,11 +84,6 @@ gistvacuumcleanup(PG_FUNCTION_ARGS)
 			totFreePages++;
 			RecordFreeIndexPage(rel, blkno);
 		}
-		else if (GistPageIsLeaf(page))
-		{
-			/* count tuples in index (considering only leaf tuples) */
-			tuplesCount += PageGetMaxOffsetNumber(page);
-		}
 		UnlockReleaseBuffer(buffer);
 	}
 
@@ -94,8 +97,6 @@ gistvacuumcleanup(PG_FUNCTION_ARGS)
 	stats->num_pages = RelationGetNumberOfBlocks(rel);
 	if (needLock)
 		UnlockRelationForExtension(rel, ExclusiveLock);
-	stats->num_index_tuples = tuplesCount;
-	stats->estimated_count = false;
 
 	PG_RETURN_POINTER(stats);
 }
