@@ -1161,11 +1161,11 @@ DecodeNumberField(int len, char *str, int fmask,
 
 		/*
 		 * OK, we have at most six digits to care about. Let's construct a
-		 * string with those digits, zero-padded on the right, and then do
-		 * the conversion to an integer.
+		 * string with those digits, zero-padded on the right, and then do the
+		 * conversion to an integer.
 		 *
-		 * XXX This truncates the seventh digit, unlike rounding it as do
-		 * the backend and the !HAVE_INT64_TIMESTAMP case.
+		 * XXX This truncates the seventh digit, unlike rounding it as do the
+		 * backend and the !HAVE_INT64_TIMESTAMP case.
 		 */
 		for (i = 0; i < 6; i++)
 			fstr[i] = *cp != '\0' ? *cp++ : '0';
@@ -2597,9 +2597,6 @@ pgtypes_defmt_scan(union un_fmt_comb * scan_val, int scan_type, char **pstr, cha
 }
 
 /* XXX range checking */
-int PGTYPEStimestamp_defmt_scan(char **, char *, timestamp *, int *, int *, int *,
-							int *, int *, int *, int *);
-
 int
 PGTYPEStimestamp_defmt_scan(char **str, char *fmt, timestamp * d,
 							int *year, int *month, int *day,
