@@ -90,6 +90,7 @@ sub DeterminePlatform
 sub IsNewer
 {
 	my ($newfile, $oldfile) = @_;
+	-e $oldfile or warn "source file \"$oldfile\" does not exist";
 	if (   $oldfile ne 'src\tools\msvc\config.pl'
 		&& $oldfile ne 'src\tools\msvc\config_default.pl')
 	{
@@ -308,7 +309,7 @@ s{PG_VERSION_STR "[^"]+"}{__STRINGIFY(x) #x\n#define __STRINGIFY2(z) __STRINGIFY
 	if ($self->{options}->{python}
 		&& IsNewer(
 			'src\pl\plpython\spiexceptions.h',
-			'src\include\backend\errcodes.txt'))
+			'src\backend\utils\errcodes.txt'))
 	{
 		print "Generating spiexceptions.h...\n";
 		system(
@@ -396,7 +397,7 @@ s{PG_VERSION_STR "[^"]+"}{__STRINGIFY(x) #x\n#define __STRINGIFY2(z) __STRINGIFY
 		  || confess "Could not open ecpg_config.h";
 		print O <<EOF;
 #if (_MSC_VER > 1200)
-#define HAVE_LONG_LONG_INT_64
+#define HAVE_LONG_LONG_INT_64 1
 #define ENABLE_THREAD_SAFETY 1
 EOF
 		print O "#define USE_INTEGER_DATETIMES 1\n"
@@ -509,10 +510,12 @@ sub AddProject
 		}
 		else
 		{
+			# We don't expect the config-specific library to be here,
+			# so don't ask for it in last parameter
 			$proj->AddLibrary(
-				$self->{options}->{openssl} . '\lib\ssleay32.lib', 1);
+				$self->{options}->{openssl} . '\lib\ssleay32.lib', 0);
 			$proj->AddLibrary(
-				$self->{options}->{openssl} . '\lib\libeay32.lib', 1);
+				$self->{options}->{openssl} . '\lib\libeay32.lib', 0);
 		}
 	}
 	if ($self->{options}->{nls})
